'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var runtimeCore = require('@vue/runtime-core');
var core = require('@nativescript/core');
var shared = require('@nativescript-vue/shared');
var set = require('set-value');
var shared$1 = require('@vue/shared');
var cssTreeParser = require('@nativescript/core/css/css-tree-parser');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e['default'] : e; }

var set__default = /*#__PURE__*/_interopDefaultLegacy(set);

const actionBarNodeOps = {
    insert(child, parent, atIndex) {
        const actionBar = parent.nativeView;
        const childView = child.nativeView;
        if (childView instanceof core.NavigationButton) {
            actionBar.navigationButton = childView;
        }
        else if (childView instanceof core.ActionItem) {
            if (atIndex) {
                const ai = actionBar.actionItems.getItems();
                ai.splice(atIndex, 0, childView);
                actionBar.actionItems.setItems(ai);
            }
            else {
                actionBar.actionItems.addItem(childView);
            }
        }
        else {
            actionBar.titleView = childView;
        }
    },
    remove(child, parent) {
        const actionBar = parent.nativeView;
        const childView = child.nativeView;
        if (childView instanceof core.NavigationButton) {
            if (actionBar.navigationButton === childView) {
                // @ts-ignore
                actionBar.navigationButton = null;
            }
        }
        else if (childView instanceof core.ActionItem) {
            actionBar.actionItems.removeItem(childView);
        }
        else {
            if (actionBar.titleView === childView) {
                // @ts-ignore
                actionBar.titleView = null;
            }
        }
    }
};
const ActionBar = (props, ctx) => {
    return runtimeCore.h('InternalActionBar', {
        ...ctx.attrs,
        onVnodeMounted(vnode) {
            const el = vnode.el; // TODO: vue-next should export ElementRender
            const actionBar = el.nativeView;
            const parent = el.parentNode;
            if (parent.nativeView instanceof core.Page) {
                parent.nativeView.actionBar = actionBar;
            }
        }
    }, ctx.slots.default ? ctx.slots.default() : undefined);
};

let defaultViewMeta = {
    viewFlags: 0 /* NONE */,
};
let elementMap = {};
function getViewMeta(elementName) {
    // console.log(`->getViewMeta(${elementName})`)
    const normalizedName = normalizeElementName(elementName);
    const entry = elementMap[normalizedName];
    if (!entry) {
        throw new Error(`No known component for element ${elementName}.`);
    }
    return entry.meta;
}
function getViewClass(elementName) {
    // console.log(`->getViewClass(${elementName})`)
    const normalizedName = normalizeElementName(elementName);
    const entry = elementMap[normalizedName];
    if (!entry) {
        throw new Error(`No known component for element ${elementName}.`);
    }
    try {
        return entry.resolver();
    }
    catch (e) {
        throw new Error(`Could not load view for: ${elementName}. ${e}`);
    }
}
function normalizeElementName(elementName) {
    return elementName.replace(/-/g, '').toLowerCase();
}
function registerElement(elementName, resolver, meta) {
    const normalizedName = normalizeElementName(elementName);
    const mergedMeta = Object.assign({}, defaultViewMeta, meta);
    if (elementMap[normalizedName] && !mergedMeta.overwriteExisting) {
        throw new Error(`Element for ${elementName} already registered.\n` +
            `If this is intentional set 'overwriteExisting: true' in 'meta'`);
    }
    elementMap[normalizedName] = {
        meta: mergedMeta,
        resolver,
    };
    // console.log(`->registerElement(${elementName})`)
}
function isKnownView(elementName) {
    return elementMap.hasOwnProperty(normalizeElementName(elementName));
}
// register built in elements
// prettier-ignore
{
    // layouts
    registerElement('AbsoluteLayout', () => require('@nativescript/core').AbsoluteLayout, { viewFlags: 4 /* LAYOUT_VIEW */ });
    registerElement('DockLayout', () => require('@nativescript/core').DockLayout, { viewFlags: 4 /* LAYOUT_VIEW */ });
    registerElement('FlexboxLayout', () => require('@nativescript/core').FlexboxLayout, { viewFlags: 4 /* LAYOUT_VIEW */ });
    registerElement('GridLayout', () => require('@nativescript/core').GridLayout, { viewFlags: 4 /* LAYOUT_VIEW */ });
    registerElement('StackLayout', () => require('@nativescript/core').StackLayout, { viewFlags: 4 /* LAYOUT_VIEW */ });
    registerElement('WrapLayout', () => require('@nativescript/core').WrapLayout, { viewFlags: 4 /* LAYOUT_VIEW */ });
    // ContentViews
    registerElement('ContentView', () => require('@nativescript/core').ContentView, { viewFlags: 2 /* CONTENT_VIEW */ });
    registerElement('ScrollView', () => require('@nativescript/core').ScrollView, { viewFlags: 2 /* CONTENT_VIEW */ });
    // ActionBar
    registerElement('InternalActionBar', () => require('@nativescript/core').ActionBar, {
        viewFlags: 1 /* SKIP_ADD_TO_DOM */,
        nodeOps: actionBarNodeOps
    });
    registerElement('ActionItem', () => require('@nativescript/core').ActionItem);
    registerElement('NavigationButton', () => require('@nativescript/core').NavigationButton);
    // navigation
    registerElement('Frame', () => require('@nativescript/core').Frame, {
        // todo: move into Frame.ts when we end up creating a component for Frame
        nodeOps: {
            insert(child, parent, atIndex) {
                const frame = parent.nativeView;
                if (child.nativeView instanceof core.Page) {
                    frame.navigate({
                        create() {
                            return child.nativeView;
                        }
                    });
                }
            },
            remove(child, parent) {
                // ignore? warn? throw? navigate back?
            }
        }
    });
    registerElement('Page', () => require('@nativescript/core').Page, { viewFlags: 2 /* CONTENT_VIEW */ });
    // html
    registerElement('HtmlView', () => require('@nativescript/core').HtmlView);
    registerElement('WebView', () => require('@nativescript/core').WebView);
    // components
    registerElement('ActivityIndicator', () => require('@nativescript/core').ActivityIndicator);
    registerElement('Button', () => require('@nativescript/core').Button);
    registerElement('DatePicker', () => require('@nativescript/core').DatePicker, {
        model: {
            prop: 'date',
            event: 'dateChange'
        }
    });
    registerElement('FormattedString', () => require('@nativescript/core').FormattedString, {
        nodeOps: {
            insert(child, parent, atIndex) {
                if (atIndex) {
                    parent.nativeView.spans.splice(atIndex, 0, child.nativeView);
                    return;
                }
                parent.nativeView.spans.push(child.nativeView);
            },
            remove(child, parent) {
                const index = parent.nativeView.spans.indexOf(child.nativeView);
                if (index > -1) {
                    parent.nativeView.spans.splice(index, 1);
                }
            }
        }
    });
    registerElement('Image', () => require('@nativescript/core').Image);
    registerElement('Label', () => require('@nativescript/core').Label);
    registerElement('ListPicker', () => require('@nativescript/core').ListPicker, {
        model: {
            prop: 'selectedIndex',
            event: 'selectedIndexChange'
        }
    });
    registerElement('InternalListView', () => require('@nativescript/core').ListView, {
        viewFlags: 8 /* NO_CHILDREN */
    });
    registerElement('Placeholder', () => require('@nativescript/core').Placeholder);
    registerElement('Progress', () => require('@nativescript/core').Progress);
    registerElement('SearchBar', () => require('@nativescript/core').SearchBar, {
        model: {
            prop: 'text',
            event: 'textChange',
        }
    });
    registerElement('SegmentedBar', () => require('@nativescript/core').SegmentedBar, {
        model: {
            prop: 'selectedIndex',
            event: 'selectedIndexChange',
        }
    });
    registerElement('SegmentedBarItem', () => require('@nativescript/core').SegmentedBarItem);
    registerElement('Slider', () => require('@nativescript/core').Slider, {
        model: {
            prop: 'value',
            event: 'valueChange',
        }
    });
    registerElement('Span', () => require('@nativescript/core').Span);
    registerElement('Switch', () => require('@nativescript/core').Switch, {
        model: {
            prop: 'checked',
            event: 'checkedChange',
        }
    });
    registerElement('TextField', () => require('@nativescript/core').TextField, {
        model: {
            prop: 'text',
            event: 'textChange',
        }
    });
    registerElement('TextView', () => require('@nativescript/core').TextView, {
        model: {
            prop: 'text',
            event: 'textChange',
        },
    });
    registerElement('TimePicker', () => require('@nativescript/core').TimePicker, {
        model: {
            prop: 'time',
            event: 'timeChange',
        },
    });
    // Tabs & BottomNavigation
    registerElement('InternalBottomNavigation', () => require('@nativescript/core').BottomNavigation, {
        model: {
            prop: 'selectedIndex',
            event: 'selectedIndexChange'
        }
    });
    registerElement('InternalTabs', () => require('@nativescript/core').Tabs, {
        model: {
            prop: 'selectedIndex',
            event: 'selectedIndexChange'
        }
    });
    registerElement('TabStrip', () => require('@nativescript/core').TabStrip, {});
    registerElement('TabStripItem', () => require('@nativescript/core').TabStripItem, {});
    registerElement('TabContentItem', () => require('@nativescript/core').TabContentItem, {});
}
// todo: more

let nodeId = 0;
class NSVNode {
    constructor(nodeType) {
        this.parentNode = null;
        this.childNodes = [];
        this.nextSibling = null;
        this.prevSibling = null;
        this.nodeType = nodeType;
        this.nodeId = nodeId++;
    }
    get text() {
        return this._text;
    }
    set text(value) {
        this._text = value;
    }
    get firstChild() {
        return this.childNodes.length ? this.childNodes[0] : null;
    }
    get lastChild() {
        return this.childNodes.length
            ? this.childNodes[this.childNodes.length - 1]
            : null;
    }
}
class NSVElement extends NSVNode {
    constructor(tagName) {
        super("element" /* ELEMENT */);
        this._tagName = normalizeElementName(tagName);
        const viewClass = getViewClass(tagName);
        this._nativeView = new viewClass();
        this._nativeView[ELEMENT_REF] = this;
    }
    get tagName() {
        return this._tagName;
    }
    get nativeView() {
        return this._nativeView;
    }
    get style() {
        return this.nativeView.style;
    }
    set style(inlineStyle) {
        this.nativeView.style = inlineStyle;
    }
    get text() {
        return this.nativeView.text;
    }
    set text(t) {
        this.nativeView.text = t;
    }
    get meta() {
        if (this._meta) {
            return this._meta;
        }
        return (this._meta = getViewMeta(this.tagName));
    }
    addEventListener(event, handler, options = {}) {
        const { capture, once } = options;
        if (capture) {
            shared.debug('Bubble propagation is not supported');
            return;
        }
        if (once) {
            const oldHandler = handler;
            const self = this;
            handler = (...args) => {
                const res = oldHandler.call(null, ...args);
                if (res !== null) {
                    self.removeEventListener(event, handler);
                }
            };
        }
        this.nativeView.addEventListener(event, handler);
    }
    removeEventListener(event, handler) {
        this.nativeView.removeEventListener(event, handler);
    }
    dispatchEvent(event) {
        this.nativeView.notify({ eventName: event, object: this.nativeView });
    }
    getAttribute(name) {
        return this.nativeView[name];
    }
    setAttribute(name, value) {
        set__default(this.nativeView, name, value);
    }
    removeAttribute(name) {
        // potential issue: unsetValue is an empty object
        // not all properties/attributes may know/check for this
        set__default(this.nativeView, name, core.unsetValue);
        // originally we deleted the property, but in case of built-in properties
        // this would break them. For example, deleting the padding property
        // will prevent us from changing the padding once we deleted it
        // that's not the expected behaviour.
        // unset(this.nativeView, name)
    }
    insertBefore(el, anchor) {
        if (!anchor) {
            return this.appendChild(el);
        }
        const refIndex = this.childNodes.findIndex((node) => node.nodeId === anchor.nodeId);
        if (refIndex === -1) {
            return this.appendChild(el);
        }
        if (el.parentNode) {
            el.parentNode.removeChild(el);
        }
        this.childNodes.splice(refIndex, 0, el);
        el.parentNode = this;
        // find index to use for the native view, since non-visual nodes
        // (comment/text don't exist in the native view hierarchy)
        // todo: potentially refactor based on my benchmark:
        // https://www.measurethat.net/Benchmarks/Show/7450/0/filter-findindex
        const trueIndex = this.childNodes
            .filter((node) => node.nodeType === "element" /* ELEMENT */)
            .findIndex((node) => node.nodeId === el.nodeId);
        this.addChild(el, trueIndex);
    }
    appendChild(el) {
        this.childNodes.push(el);
        el.parentNode = this;
        this.addChild(el);
    }
    removeChild(el) {
        const index = this.childNodes.findIndex((node) => node.nodeId === el.nodeId);
        if (index > -1) {
            this.childNodes.splice(index, 1);
            el.parentNode = null;
            if (el.nodeType === "element" /* ELEMENT */) {
                removeChild(el, this);
            }
            else if (el.nodeType === "text" /* TEXT */) {
                this.updateText();
            }
        }
    }
    // abstracted from appendChild, and insertBefore to avoid code duplication
    addChild(el, atIndex) {
        if (el.nodeType === "element" /* ELEMENT */) {
            addChild(el, this, atIndex);
        }
        else if (el.nodeType === "text" /* TEXT */) {
            this.updateText();
        }
    }
    updateText() {
        this.setAttribute('text', this.childNodes
            .filter((node) => node.nodeType === "text" /* TEXT */)
            .reduce((text, currentNode) => {
            return text + currentNode.text;
        }, ''));
    }
}
class NSVComment extends NSVNode {
    constructor(text) {
        super("comment" /* COMMENT */);
        this.text = text;
    }
}
class NSVText extends NSVNode {
    constructor(text) {
        super("text" /* TEXT */);
        this.text = text;
    }
}
class NSVRoot extends NSVNode {
    constructor() {
        super("root" /* ROOT */);
    }
    appendChild(el) {
        // console.log(`NSVRoot->appendChild(${el.nodeType})`)
        if (el instanceof NSVElement) {
            this.el = el;
        }
        // no-op
    }
}
function addChild(child, parent, atIndex) {
    // debug(
    //   `...addChild(  ${child.tagName}(${child.nodeId}), ${parent.tagName}(${
    //     parent.nodeId
    //   }), ${atIndex}  )`
    // )
    if (child.meta.viewFlags & 1 /* SKIP_ADD_TO_DOM */) {
        // debug('SKIP_ADD_TO_DOM')
        return;
    }
    const parentView = parent.nativeView;
    const childView = child.nativeView;
    if (parent.meta.viewFlags & 8 /* NO_CHILDREN */) {
        // debug('NO_CHILDREN')
        return;
    }
    if (parent.meta.nodeOps) {
        return parent.meta.nodeOps.insert(child, parent, atIndex);
    }
    if (parent.meta.viewFlags & 4 /* LAYOUT_VIEW */) {
        if (atIndex) {
            parentView.insertChild(childView, atIndex);
        }
        else {
            parentView.addChild(childView);
        }
    }
    else if (parent.meta.viewFlags & 2 /* CONTENT_VIEW */) {
        parentView.content = childView;
    }
    else {
        parentView._addChildFromBuilder(childView.constructor.name, childView);
    }
}
function removeChild(child, parent) {
    // debug(
    //   `...removeChild(  ${child.tagName}(${child.nodeId}), ${parent.tagName}(${
    //     parent.nodeId
    //   })  )`
    // )
    if (child.meta.viewFlags & 1 /* SKIP_ADD_TO_DOM */) {
        // debug('SKIP_ADD_TO_DOM')
        return;
    }
    if (parent.meta.viewFlags & 8 /* NO_CHILDREN */) {
        // debug('NO_CHILDREN')
        return;
    }
    if (parent.meta.nodeOps) {
        return parent.meta.nodeOps.remove(child, parent);
    }
    const parentView = parent.nativeView;
    const childView = child.nativeView;
    if (parent.meta.viewFlags & 4 /* LAYOUT_VIEW */) {
        parentView.removeChild(childView);
    }
    else if (parent.meta.viewFlags & 2 /* CONTENT_VIEW */) {
        parentView.content = null;
    }
    else {
        parentView._removeView(childView);
    }
}

const nodeOps = {
    createRoot() {
        return new NSVRoot();
    },
    createComment(text) {
        return new NSVComment(text);
    },
    createElement(type, isSVG) {
        return new NSVElement(type);
    },
    createText(text) {
        return new NSVText(text);
    },
    nextSibling(node) {
        return node.nextSibling;
    },
    parentNode(node) {
        return node.parentNode;
    },
    insert(child, parent, anchor) {
        if (anchor !== null) {
            parent.insertBefore(child, anchor);
        }
        else {
            parent.appendChild(child);
        }
    },
    remove(el) {
        if (el.parentNode != null) {
            el.parentNode.removeChild(el);
        }
    },
    setElementText(node, text) {
        node.text = text;
    },
    setText(node, text) {
        node.text = text;
    },
    setScopeId(el, id) {
        el.setAttribute(id, '');
    }
};

function addEventListener(el, event, handler, options = {}) {
    el.addEventListener(event, handler, options);
}
function patchEvent(el, name, prevValue, nextValue) {
    const prevOptions = prevValue && 'options' in prevValue && prevValue.options;
    const nextOptions = nextValue && 'options' in nextValue && nextValue.options;
    const invoker = prevValue && prevValue.invoker;
    const value = nextValue && 'handler' in nextValue ? nextValue.handler : nextValue;
    if (prevOptions || nextOptions) {
        const prev = prevOptions || shared$1.EMPTY_OBJ;
        const next = nextOptions || shared$1.EMPTY_OBJ;
        if (prev.once !== next.once) {
            if (invoker) {
                el.removeEventListener(name, invoker);
            }
            if (nextValue && value) {
                const invoker = createInvoker(value);
                nextValue.invoker = invoker;
                addEventListener(el, name, invoker, next);
            }
            return;
        }
        return;
    }
    if (nextValue && value) {
        if (invoker) {
            prevValue.invoker = null;
            invoker.value = value;
            nextValue.invoker = invoker;
        }
        else {
            addEventListener(el, name, createInvoker(value));
        }
    }
    else if (invoker) {
        el.removeEventListener(name, invoker);
    }
}
function createInvoker(initialValue) {
    const invoker = (e) => {
        runtimeCore.callWithAsyncErrorHandling(invoker.value, null, 5, [e]);
    };
    invoker.value = initialValue;
    initialValue.invoker = invoker;
    return invoker;
}

// compiler should normalize class + :class bindings on the same element
// into a single binding ['staticClass', dynamic]
function patchClass(el, value) {
    if (value == null) {
        value = '';
    }
    el.setAttribute('class', value);
}

function isRule(node) {
    return node.type === 'rule';
}
function isDeclaration(node) {
    return node.type === 'declaration';
}
function createDeclaration(decl) {
    return { property: decl.property.toLowerCase(), value: decl.value };
}
function declarationsFromAstNodes(astRules) {
    return astRules.filter(isRule).map((rule) => {
        return rule.declarations.filter(isDeclaration).map(createDeclaration);
    });
}
function patchStyle(el, prev, next) {
    if (prev) {
        // todo: check if we can substitute cssTreeParse with parseInlineCSS from compiler/transforms/transformStyle (by extracting it to shared)
        // reset previous styles
        const localStyle = `local { ${prev} }`;
        const ast = cssTreeParser.cssTreeParse(localStyle, undefined);
        const [declarations] = declarationsFromAstNodes(ast.stylesheet.rules);
        declarations.forEach((d) => {
            el.nativeView.style[d.property] = core.unsetValue;
        });
    }
    if (!next) {
        el.removeAttribute('style');
    }
    else {
        // set new styles
        el.style = next;
    }
}

function patchAttr(el, key, prevValue, nextValue) {
    if (isAndroidKey(key)) {
        if (!core.isAndroid) {
            // if we encounter an android key, and we are not on android we can safely ignore it
            return;
        }
        key = key.substr(8);
    }
    else if (isIOSKey(key)) {
        if (!core.isIOS) {
            // if we encounter an ios key, and we are not on ios we can safely ignore it
            return;
        }
        key = key.substr(4);
    }
    // detect expandable attrs for boolean values
    // see https://vuejs.org/v2/guide/components-props.html#Passing-a-Boolean
    if (isBoolean(el.getAttribute(key)) && nextValue === '') {
        nextValue = true;
    }
    if (nextValue === null) {
        // if the nextValue is null, we are removing the attribute
        el.removeAttribute(key);
    }
    else {
        el.setAttribute(key, nextValue);
    }
}

function patchProp(el, key, prevValue, nextValue) {
    // console.log('->patchProp')
    switch (key) {
        // special
        case 'class':
            // console.log('->patchProp+Class')
            patchClass(el, nextValue);
            break;
        case 'style':
            // console.log('->patchProp+Style')
            patchStyle(el, prevValue, nextValue);
            break;
        case 'modelValue':
        case 'onUpdate:modelValue':
            // handled by v-model directive
            break;
        default:
            if (shared.isOn(key)) {
                patchEvent(el, key.slice(2), prevValue, nextValue);
            }
            else {
                patchAttr(el, key, prevValue, nextValue);
            }
    }
}

/**
 * @internal
 */
function install(app) {
    app.config.globalProperties.$navigateTo = $navigateTo;
    app.config.globalProperties.$navigateBack = $navigateBack;
}
function resolveFrame(frame) {
    if (!frame) {
        return core.Frame.topmost();
    }
    if (frame instanceof core.Frame) {
        return frame;
    }
    // todo: check if refs work this way or not
    if (runtimeCore.isRef(frame)) {
        return frame.value;
    }
    if (frame instanceof NSVElement) {
        return frame.nativeView;
    }
    // todo: either change core Frame to add frames to the stack when they are created
    // or do as we did in 2.x - handle a Map of frames.
    // right now, empty frames can't be navigated as they are not recognized by `getFrameById`
    return core.Frame.getFrameById(frame);
}
async function $navigateTo(target, options) {
    options = Object.assign({}, options);
    console.log('$navigateTo');
    try {
        const frame = resolveFrame(options.frame);
        if (!frame) {
            throw new Error('Failed to resolve frame. Make sure your frame exists.');
        }
        const navigationApp = createApp(target, options.props);
        const targetPage = navigationApp.mount().$el.nativeView;
        const handler = (args) => {
            if (args.isBackNavigation) {
                targetPage.off('navigatedFrom', handler);
                navigationApp.unmount();
            }
        };
        targetPage.on('navigatedFrom', handler);
        const dispose = targetPage.disposeNativeView;
        targetPage.disposeNativeView = () => {
            navigationApp.unmount();
            dispose.call(targetPage);
        };
        frame.navigate({
            ...options,
            create: () => targetPage,
        });
        return targetPage;
    }
    catch (e) {
        console.log('[$navigateTo] Failed to navigate:\n\n');
        console.log(e);
        throw e;
    }
}
async function $navigateBack(options) {
    const frame = resolveFrame(options ? options.frame : undefined);
    if (!frame) {
        throw new Error('Failed to resolve frame. Make sure your frame exists.');
    }
    if (!frame.canGoBack()) {
        return;
    }
    frame.goBack();
}

/**
 * @internal
 */
function install$1(app) {
    app.config.globalProperties.$showModal = $showModal;
}
function resolveModalTarget(target) {
    if (shared$1.isObject(target) && shared$1.isObject(target.$el)) {
        return target.$el.nativeView;
    }
    else if (target instanceof NSVElement) {
        return target.nativeView;
    }
    else if (target instanceof core.View) {
        return target;
    }
    return false;
}
async function $showModal(component, options) {
    options = Object.assign({
        target: core.Application.getRootView(),
    }, options);
    const modalTarget = resolveModalTarget(options.target);
    if (!modalTarget) {
        return false;
    }
    return new Promise((resolve) => {
        const modalApp = createApp(component, options ? options.props : null);
        let isResolved = false;
        const closeCallback = (data) => {
            if (isResolved)
                return;
            isResolved = true;
            try {
                modalContent.closeModal();
            }
            catch (e) {
                // ignore?
            }
            modalApp.unmount();
            resolve(data);
        };
        modalApp.config.globalProperties.$modal = {
            close: closeCallback,
        };
        const modalContent = modalApp.mount().$el.nativeView;
        modalTarget.showModal(modalContent, {
            ...options,
            context: null,
            closeCallback,
        });
    });
}

const ELEMENT_REF = Symbol( ``);
const isAndroidKey = (key) => key[0] === 'a' &&
    key[1] === 'n' &&
    key[2] === 'd' &&
    key[3] === 'r' &&
    key[4] === 'o' &&
    key[5] === 'i' &&
    key[6] === 'd' &&
    key[7] === ':';
const isIOSKey = (key) => key[0] === 'i' && key[1] === 'o' && key[2] === 's' && key[3] === ':';
const isBoolean = (value) => {
    return typeof value === 'boolean' || value instanceof Boolean;
};

function resolveComponent(name) {
    // in the standalone compiler, everything is treated as a component because we don't
    // know if certain tags are elements or not at runtime
    // (they are only registered at runtime with registerElement)
    // if we return a string here, vue will render them as normal elements
    // with the default slot as children
    if (isKnownView(name)) {
        return name;
    }
    // todo: refactor to not have to hardcode all built-in components
    if (name === 'ActionBar') {
        return ActionBar;
    }
    if (name === 'Tabs') {
        return Tabs;
    }
    if (name === 'BottomNavigation') {
        return BottomNavigation;
    }
    return runtimeCore.resolveComponent(name);
}

const getModelAssigner = (vnode) => {
    const fn = vnode.props['onUpdate:modelValue'];
    return shared$1.isArray(fn) ? (value) => shared$1.invokeArrayFns(fn, value) : fn;
};
function toNumber(val) {
    const n = parseFloat(val);
    return isNaN(n) ? val : n;
}
const vModel = {
    beforeMount(el, { value, modifiers: { trim, number } }, vnode) {
        el._assign = getModelAssigner(vnode);
        const castToNumber = number;
        const { prop, event } = getViewMeta(el.tagName).model;
        el._assign(value);
        el.setAttribute(prop, value);
        addEventListener(el, event, () => {
            let propValue = el.getAttribute(prop);
            if (trim && typeof propValue === 'string') {
                propValue = propValue.trim();
            }
            else if (castToNumber && typeof propValue === 'string') {
                propValue = toNumber(propValue);
            }
            el._assign(propValue);
        });
    },
    beforeUpdate(el, { value, oldValue }, vnode) {
        const { prop } = getViewMeta(el.tagName).model;
        el._assign = getModelAssigner(vnode);
        if (value === oldValue) {
            return;
        }
        el.setAttribute(prop, value);
    },
};

const vShow = {
    beforeMount(el, { value }, { transition }) {
        el._vod =
            el.getAttribute('visibility') === 'none'
                ? ''
                : el.getAttribute('visibility');
        if (transition && value) {
            transition.beforeEnter(el);
        }
        else {
            setDisplay(el, value);
        }
    },
    mounted(el, { value }, { transition }) {
        if (transition && value) {
            transition.enter(el);
        }
    },
    updated(el, { value, oldValue }, { transition }) {
        if (!value === !oldValue)
            return;
        if (transition) {
            if (value) {
                transition.beforeEnter(el);
                setDisplay(el, true);
                transition.enter(el);
            }
            else {
                transition.leave(el, () => {
                    setDisplay(el, false);
                });
            }
        }
        else {
            setDisplay(el, value);
        }
    },
    beforeUnmount(el) {
        setDisplay(el, true);
    }
};
function setDisplay(el, value) {
    el.setAttribute('visibility', value ? el._vod : 'collapsed');
}

const NSV_LVR_REF = Symbol('NSV_LVR_REF');
const isListViewThingSymbol = Symbol();
// TODO: rewrite without the keyedTemplate hack, as it will not work for recycler views
const ListView = (props, ctx) => {
    const defaultKeyedTemplate = {
        key: 'test',
        createView() {
            // debug('createView for test template')
            return {
                [isListViewThingSymbol]: true,
                container: null,
            };
        },
    };
    return runtimeCore.h('InternalListView', {
        items: props.items,
        onitemTap: (args) => {
            ctx.emit('itemTap', {
                ...args,
                item: props.items[args.index],
            });
        },
        itemTemplates: [defaultKeyedTemplate],
        itemTemplateSelector: (item, index) => {
            // debug('itemTemplateSelector', item, index)
            return 'test';
        },
        onitemLoading(args) {
            // debug('item loading', args.view)
            const item = props.items[args.index];
            const getSlotVnode = (slotName) => ctx.slots[slotName]({
                item,
                index: args.index,
                even: !!(args.index % 2),
                odd: !(args.index % 2),
            })[0];
            let view;
            if (args.view[isListViewThingSymbol]) {
                shared.debug('INITIAL LVR CREATING' + args.index, 'ListView');
                // it is an initial render into this view holder
                const lvr = args.view;
                // lets create an initial root
                lvr.container = nodeOps.createRoot();
                // @ts-ignore
                // console.log(getSlotVnode('default').children[0].children)
                render(getSlotVnode('default'), lvr.container);
                if (lvr.container.el) {
                    lvr.container.el.nativeView[NSV_LVR_REF] = lvr;
                    view = lvr.container.el.nativeView;
                }
            }
            else {
                shared.debug('LVR UPDATING' + args.index, 'ListView');
                const lvr = args.view[NSV_LVR_REF];
                // @ts-ignore
                // console.log(getSlotVnode('default').children[0].children)
                render(getSlotVnode('default'), lvr.container);
                view = lvr.container.el.nativeView;
            }
            args.view = view;
        },
    });
};

const Tabs = (props, ctx) => {
    return runtimeCore.h('InternalTabs', {
        ...ctx.attrs,
        onVnodeMounted(vnode) {
            const el = vnode.el;
            console.log(el);
            // TODO: Implement
        },
    }, ctx.slots.default ? ctx.slots.default() : undefined);
};

const BottomNavigation = (props, ctx) => {
    return runtimeCore.h('InternalBottomNavigation', {
        ...ctx.attrs,
        onVnodeMounted(vnode) {
            const el = vnode.el;
            console.log(el);
            // TODO: Implement
        },
    }, ctx.slots.default ? ctx.slots.default() : undefined);
};

const rendererOptions = {
    patchProp,
    ...nodeOps,
};
let renderer;
function ensureRenderer() {
    return (renderer ||
        (renderer = runtimeCore.createRenderer(rendererOptions)));
}
function runApp(root) {
    core.Application.run({
        create: () => {
            shared.debug(`Root Node: ${JSON.stringify({
                id: root.$el.nodeId,
                type: root.$el.nodeType,
                tag: root.$el.tagName,
            })}`);
            return root.$el.nativeView;
        },
    });
    return root;
}
const render = ((...args) => {
    ensureRenderer().render(...args);
});
const createApp = ((...args) => {
    const app = ensureRenderer().createApp(...args);
    const { mount, unmount } = app;
    const root = nodeOps.createRoot();
    app.mount = () => {
        return mount(root);
    };
    app.unmount = () => {
        return unmount(root);
    };
    app.start = () => {
        return runApp(app.mount());
    };
    // Built-in plugins
    app.use(install);
    app.use(install$1);
    return app;
});

Object.keys(runtimeCore).forEach(function (k) {
  if (k !== 'default') exports[k] = runtimeCore[k];
});
exports.$navigateTo = $navigateTo;
exports.$showModal = $showModal;
exports.ActionBar = ActionBar;
exports.BottomNavigation = BottomNavigation;
exports.ELEMENT_REF = ELEMENT_REF;
exports.ListView = ListView;
exports.NSVComment = NSVComment;
exports.NSVElement = NSVElement;
exports.NSVNode = NSVNode;
exports.NSVRoot = NSVRoot;
exports.NSVText = NSVText;
exports.Tabs = Tabs;
exports.createApp = createApp;
exports.defaultViewMeta = defaultViewMeta;
exports.getViewClass = getViewClass;
exports.getViewMeta = getViewMeta;
exports.isAndroidKey = isAndroidKey;
exports.isBoolean = isBoolean;
exports.isIOSKey = isIOSKey;
exports.isKnownView = isKnownView;
exports.nodeOps = nodeOps;
exports.normalizeElementName = normalizeElementName;
exports.registerElement = registerElement;
exports.render = render;
exports.resolveComponent = resolveComponent;
exports.vModel = vModel;
exports.vShow = vShow;
